#include <EEPROM.h>
#include <LiquidCrystal.h>
#include "SparkFunSi4703.h"

/*
|1 |0|3|.|8|M|H|z| | |S|t|e|r|e|o|
|Ω|1|6| |S|e|e|k| |>|7|5|d|B|u|V|
*/

//Though it's not recommended to use macros in C++, I still prefer using macros rather than constants.

#define FREQ_STORED_ADDRU 0                            //The address of the upper 8 bits of frequency value stored in EEPROM. The frequency value is the actual frequency multiplied by 10.
#define FREQ_STORED_ADDRL 1                            //The address of the lower 8 bits of frequency value stored in EEPROM. 
#define VOL_STORED_ADDR   2                            //The address of the volume stored in EEPROM.

#define DEFAULT_FREQ   1068
#define DEFAULT_VOL   2

#define KEY_UP 10                                      //The "UP" key
#define KEY_DOWN 8
#define KEY_FUNC 9                                     //The "Function Switch" key

#define TUNE_COARSE 0                                  //Coarse-Tuning
#define TUNE_FINE 1                                    //Fine-Tuning
#define VOL_ADJ 2                                      //Volume Adjustment 
#define SEEK 3

#define CHK_KEY_STATE_DLY 200
#define KEY_PRESSED 1
#define KEY_RELEASED 0

LiquidCrystal lcd(12, 11, 5, 4, 3, 2);               //RS, EN, D4, D5, D6, D7
Si4703_Breakout radio(A0, A4, A5);                     //RST, SDIO, SCLK

unsigned char FuncSel = TUNE_COARSE;                   //Function selection
unsigned short CurrentFreq;
unsigned char CurrentVol, RSSI, IsStereo;
unsigned char RefreshCounter = 0;
unsigned char Key_State;

unsigned char Speaker[8] = {                          //Speaker symbol on LCD
  B00001,
  B00011,
  B11101,
  B10101,
  B10101,
  B11101,
  B00011,
  B00001
};

void DisplayFreq(unsigned short Freq)
{
  if(Freq >= 1000)
  {
    lcd.setCursor(0, 0);
    lcd.print(Freq/10, DEC);
    lcd.setCursor(4, 0);
    lcd.print(Freq%10, DEC);
  }
  else
  {
    lcd.setCursor(0, 0);
    lcd.print(" ");
    lcd.print(Freq/10, DEC);
    lcd.setCursor(4, 0);
    lcd.print(Freq%10, DEC);
  }
}

void DisplayVol(unsigned char Vol)
{
  lcd.setCursor(1, 1);
  lcd.print(Vol, DEC);
  if(Vol < 10)
  {
    lcd.print(" ");
  }
}

unsigned char CheckKeyState(unsigned char KeyIdx)
{
  if(digitalRead(KeyIdx) == LOW)
  {
    delay(50);
    if(digitalRead(KeyIdx) == LOW)
    {
      return KEY_PRESSED;
    }
    return KEY_RELEASED;
  }
  else if(digitalRead(KeyIdx) == HIGH)
  {
    delay(50);
    if(digitalRead(KeyIdx) == HIGH)
    {
      return KEY_RELEASED;
    }
    return KEY_PRESSED;
  }
}

void RestoreCursor()
{
  switch(FuncSel)                                                     
  {
    case TUNE_COARSE:
      lcd.setCursor(2, 0); break;
    case TUNE_FINE:
      lcd.setCursor(4, 0); break;
    case VOL_ADJ:
      lcd.setCursor(0, 1); break;
    case SEEK:
      lcd.setCursor(4, 1); break;
    default:
      FuncSel = TUNE_COARSE; lcd.setCursor(2, 0); break;
  }
}

void setup()
{
  pinMode(KEY_DOWN, INPUT);                           //K1, -
  pinMode(KEY_UP, INPUT);                             //K3, +
  pinMode(KEY_FUNC, INPUT);                           //K2, Function Switch
  pinMode(13, INPUT);                                 //Connected to GPIO2 of Si4703, reserved
  
  lcd.createChar(0, Speaker);
  lcd.begin(16, 2);
  lcd.setCursor(3, 0);
  lcd.print(". MHz");
  lcd.setCursor(0, 1);
  lcd.write(byte(0));                                 //Draw speaker symbol
  lcd.print("   Seek    dBuV");
  radio.PowerOn();
  
  unsigned char StoredVol;
  unsigned short StoredFreq;
  StoredFreq = EEPROM.read(FREQ_STORED_ADDRU);
  StoredFreq = (StoredFreq<<8) | EEPROM.read(FREQ_STORED_ADDRL);
  StoredVol = EEPROM.read(VOL_STORED_ADDR);
  if(StoredVol <= 15)                                                     //Restore volume from EEPROM after power on
  {
    CurrentVol = StoredVol;
    DisplayVol(StoredVol);
    radio.SetVolume(StoredVol);
  }
  else
  {
    DisplayVol(DEFAULT_VOL);
    radio.SetVolume(DEFAULT_VOL);
  }
  if((StoredFreq >= 760)&&(StoredFreq <= 1080))                           //Restore channel from EEPROM after power on
  {
    CurrentFreq = StoredFreq;
    DisplayFreq(StoredFreq);
    radio.SetFrequency(StoredFreq);
  }
  else
  {
    CurrentFreq = DEFAULT_FREQ;
    DisplayFreq(DEFAULT_FREQ);
    radio.SetFrequency(DEFAULT_FREQ);
  }

  lcd.setCursor(2, 0);
  lcd.cursor();
}

void loop()
{
/////////////////////////////////////Function switch routine////////////////////////////////////////////
  Key_State = CheckKeyState(KEY_FUNC);
  while(Key_State == KEY_PRESSED)
  {
    delay(CHK_KEY_STATE_DLY);
    Key_State = CheckKeyState(KEY_FUNC);
    if(Key_State == KEY_RELEASED)                                         //Toggle function only after key is released
    {
      switch(FuncSel)
      {
        case TUNE_COARSE:
          FuncSel = TUNE_FINE; lcd.setCursor(4, 0); break;
        case TUNE_FINE:
          FuncSel = VOL_ADJ; lcd.setCursor(0, 1); break;
        case VOL_ADJ:
          FuncSel = SEEK; lcd.setCursor(4, 1); break;
        case SEEK:
          FuncSel = TUNE_COARSE; lcd.setCursor(2, 0); break;
        default:
          FuncSel = TUNE_COARSE; lcd.setCursor(2, 0); break;
      }
      break;
    }
  }
////////////////////////////////////Coarse-Tuning. Change the integral part of the actual frequency value/////////////////////////////////////////////
  if(FuncSel == TUNE_COARSE)
  {
    Key_State = CheckKeyState(KEY_DOWN);
    while(Key_State == KEY_PRESSED)
    {
      if(CurrentFreq >= 770)
      {
        lcd.setCursor(10, 0);
        lcd.print("      ");                                            //Clear stereo/mono indicator on LCD when changing channel
        lcd.setCursor(10, 1);
        lcd.print(" -");                                                //Clear RSSI indicator on LCD when changing channel
        CurrentFreq -= 10;
        DisplayFreq(CurrentFreq);
        lcd.setCursor(2, 0);
      }
      delay(CHK_KEY_STATE_DLY);
      Key_State = CheckKeyState(KEY_DOWN);
      if(Key_State == KEY_RELEASED)                                     //Change channel and store the frequency value to EEPROM only after key is released
      {
        radio.SetFrequency(CurrentFreq);
        unsigned char Val;
        Val =  CurrentFreq & 0x00FF;
        EEPROM.update(FREQ_STORED_ADDRL, Val);
        Val =  CurrentFreq >> 8;
        EEPROM.update(FREQ_STORED_ADDRU, Val);
      }
    }

    Key_State = CheckKeyState(KEY_UP);
    while(Key_State == KEY_PRESSED)
    {
      if(CurrentFreq <= 1070)
      {
        lcd.setCursor(10, 0);
        lcd.print("      ");                                            //Clear stereo/mono indicator on LCD when changing channel
        lcd.setCursor(10, 1);
        lcd.print(" -");                                                //Clear RSSI indicator on LCD when changing channel
        CurrentFreq += 10;
        DisplayFreq(CurrentFreq);
        lcd.setCursor(2, 0);
      }
      delay(CHK_KEY_STATE_DLY);
      Key_State = CheckKeyState(KEY_UP);
      if(Key_State == KEY_RELEASED)                                     //Change channel and store the frequency value to EEPROM only after key is released
      {
        radio.SetFrequency(CurrentFreq);
        unsigned char Val;
        Val =  CurrentFreq & 0x00FF;
        EEPROM.update(FREQ_STORED_ADDRL, Val);
        Val =  CurrentFreq >> 8;
        EEPROM.update(FREQ_STORED_ADDRU, Val);
      }
    }
  }
////////////////////////////////////Fine-Tuning. Change the fractional part of the actual frequency value/////////////////////////////////////////////
  else if(FuncSel == TUNE_FINE)
  {
    Key_State = CheckKeyState(KEY_DOWN);
    while(Key_State == KEY_PRESSED)
    {
      if(CurrentFreq > 760)
      {
        lcd.setCursor(10, 0);
        lcd.print("      ");                                            //Clear stereo/mono indicator on LCD when changing channel
        lcd.setCursor(10, 1);
        lcd.print(" -");                                                //Clear RSSI indicator on LCD when changing channel
        CurrentFreq -= 1;
        DisplayFreq(CurrentFreq);
        lcd.setCursor(4, 0);
      }
      delay(CHK_KEY_STATE_DLY);
      Key_State = CheckKeyState(KEY_DOWN);
      if(Key_State == KEY_RELEASED)                                     //Change channel and store the frequency value to EEPROM only after key is released
      {
        radio.SetFrequency(CurrentFreq);
        unsigned char Val;
        Val =  CurrentFreq & 0x00FF;
        EEPROM.update(FREQ_STORED_ADDRL, Val);
        Val =  CurrentFreq >> 8;
        EEPROM.update(FREQ_STORED_ADDRU, Val);
      }
    }

    Key_State = CheckKeyState(KEY_UP);
    while(Key_State == KEY_PRESSED)
    {
      if(CurrentFreq < 1080)
      {
        lcd.setCursor(10, 0);
        lcd.print("      ");                                            //Clear stereo/mono indicator on LCD when changing channel
        lcd.setCursor(10, 1);
        lcd.print(" -");                                                //Clear RSSI indicator on LCD when changing channel
        CurrentFreq += 1;
        DisplayFreq(CurrentFreq);
        lcd.setCursor(4, 0);
      }
      delay(CHK_KEY_STATE_DLY);
      Key_State = CheckKeyState(KEY_UP);
      if(Key_State == KEY_RELEASED)                                     //Change channel and store the frequency value to EEPROM only after key is released
      {
        radio.SetFrequency(CurrentFreq);
        unsigned char Val;
        Val =  CurrentFreq & 0x00FF;
        EEPROM.update(FREQ_STORED_ADDRL, Val);
        Val =  CurrentFreq >> 8;
        EEPROM.update(FREQ_STORED_ADDRU, Val);
      }
    }
  }
////////////////////////////////////Volume Adjustment/////////////////////////////////////////////
  else if(FuncSel == VOL_ADJ)
  {
    Key_State = CheckKeyState(KEY_DOWN);
    while(Key_State == KEY_PRESSED)
    {
      if(CurrentVol > 0)
      {
        CurrentVol -= 1;
        DisplayVol(CurrentVol);
        lcd.setCursor(0, 1);
        radio.SetVolume(CurrentVol);
      }
      delay(CHK_KEY_STATE_DLY);
      Key_State = CheckKeyState(KEY_DOWN);
      if(Key_State == KEY_RELEASED)                                     //Store the volume to EEPROM only after key is released
      {
        EEPROM.write(VOL_STORED_ADDR, CurrentVol);
      }
    }

    Key_State = CheckKeyState(KEY_UP);
    while(Key_State == KEY_PRESSED)
    {
      if(CurrentVol < 15)
      {
        CurrentVol += 1;
        DisplayVol(CurrentVol);
        lcd.setCursor(0, 1);
        radio.SetVolume(CurrentVol);
      }
      delay(CHK_KEY_STATE_DLY);
      Key_State = CheckKeyState(KEY_UP);
      if(Key_State == KEY_RELEASED)                                     //Store the volume to EEPROM only after key is released
      {
        EEPROM.write(VOL_STORED_ADDR, CurrentVol);
      }
    }
  }
////////////////////////////////////Seek/////////////////////////////////////////////
  else if(FuncSel == SEEK)                                              //Seeking result will not be stored to EEPROM given the write/erase cycles of it
  {
    Key_State = CheckKeyState(KEY_DOWN);
    while(Key_State == KEY_PRESSED)
    {
      delay(CHK_KEY_STATE_DLY);
      Key_State = CheckKeyState(KEY_DOWN);
      if(Key_State == KEY_RELEASED)
      {
        lcd.blink();
        unsigned short SeekResult =  radio.SeekDown();
        lcd.noBlink();
        if(SeekResult != 0)
        {
          CurrentFreq = SeekResult;
          DisplayFreq(CurrentFreq);
          lcd.setCursor(4, 1);
        }//If the seek operation is unable to find a valid channel, the seek failure/band limit (SF/BL) bit is set high and the device returns to the channel selected before the seek operation began.
      }
    }
    
    Key_State = CheckKeyState(KEY_UP);
    while(Key_State == KEY_PRESSED)
    {
      delay(CHK_KEY_STATE_DLY);
      Key_State = CheckKeyState(KEY_UP);
      if(Key_State == KEY_RELEASED)
      {
        lcd.blink();
        unsigned short SeekResult =  radio.SeekUp();
        lcd.noBlink();
        if(SeekResult != 0)
        {
          CurrentFreq = SeekResult;
          DisplayFreq(CurrentFreq);
          lcd.setCursor(4, 1);
        }//If the seek operation is unable to find a valid channel, the seek failure/band limit (SF/BL) bit is set high and the device returns to the channel selected before the seek operation began.
      }
    }
  }
////////////////////////////////////Stereo/mono and RSSI indication/////////////////////////////////////////////
  if(RefreshCounter < 3)
  {
    RefreshCounter++;
  }
  else
  {
    RefreshCounter = 0;
    radio.GetStatus(&RSSI, &IsStereo, NULL);
    lcd.setCursor(9, 1);
    if(RSSI < 10)
    {
      lcd.print("  ");
    }
    else if(RSSI >= 75)
    {
      lcd.print(">");                                                   //Si4702/03-C19 does not report RSSI levels greater than 75 dBuV
    }
    else
    {
      lcd.print(" ");
    }
    lcd.print(RSSI, DEC);
  
    lcd.setCursor(10, 0);
    if(IsStereo)
    {
      lcd.print("Stereo");
    }
    else
    {
      lcd.print("Mono  ");
    }
    RestoreCursor();
  }
}
