#include "SparkFunSi4703.h"

Si4703_Breakout::Si4703_Breakout(unsigned char resetPin, unsigned char sdioPin, unsigned char sclkPin)//According to pins_arduino.h, the data type of pin numbers are uint8_t
{
  _resetPin = resetPin;
  _sdioPin = sdioPin;
  _sclkPin = sclkPin;
}

void Si4703_Breakout::PowerOn()
{
	if(IsPoweredOn)
	{
    return;
	}
	Si4703_Init();
  IsPoweredOn = true;
}

//To get the Si4703 inito 2-wire mode, SEN needs to be high and SDIO needs to be low after a reset
//The breakout board has SEN pulled high, but also has SDIO pulled high. Therefore, after a normal power up
//The Si4703 will be in an unknown state. RST must be controlled
void Si4703_Breakout::Si4703_Init() 
{
  pinMode(_resetPin, OUTPUT);
  pinMode(_sdioPin, OUTPUT);                                        //SDIO is connected to A4 for I2C
  digitalWrite(_sdioPin, LOW);                                      //A low SDIO indicates a 2-wire interface
  delay(1);                                                         //Some delays while we allow pins to settle
  digitalWrite(_resetPin, LOW);                                     //Put Si4703 into reset
  delay(1);                                                         //Some delays while we allow pins to settle
  digitalWrite(_resetPin, HIGH);                                    //Bring Si4703 out of reset with SDIO set to low and SEN pulled high with on-board resistor
  delay(1);                                                         //Allow Si4703 to come out of reset

  Wire.begin();                                                     //Now that the unit is reset and I2C inteface mode, we need to begin I2C
  Wire.setClock(400000);                                            //400kHz I2C clock frequency. Wire.setClock() must be called after Wire.begin(). Wire.begin() initializes the bitrate to 100kHz.

  ReadRegisters();                                                  //Read the current register set
  Si4703_Registers[TEST1] = 0x8100;                                 //Enable the oscillator, from AN230 page 12, rev 0.9
  UpdateRegisters();                                                //Update
  delay(600);                                                       //Wait for clock to settle - from AN230 page 12, rev 0.9

  ReadRegisters();                                                  //Read the current register set
  Si4703_Registers[POWERCFG] = 0x4001;                              //Enable the IC
  UpdateRegisters();
  delay(110);                                                       //Max powerup time is 110ms, from datasheet page 13
  
  ReadRegisters();
  Si4703_Registers[SYSCONFIG1] |= 0x1800;                           //Enable RDS. Set FM De-Emphasis to 50us
  Si4703_Registers[SYSCONFIG2] |= 0x0051;                           //76–108 MHz, 100kHz channel spacing. Set volume to lowest
  Si4703_Registers[SYSCONFIG3] |= 0x0011;                           //Seek SNR Threshold: min (most stops), Seek FM Impulse Detection Threshold: max (most stops)
  UpdateRegisters();                                                //Write remaining configuration registers
}

void Si4703_Breakout::SetFrequency(unsigned short Freq)             //Freq is the actual frequency multiplied by 10. Channel spacing is 0.1MHz.
{
  if((Freq < 760)||(Freq > 1080))
  {
    return;
  }
  unsigned short Channel;
  Channel = Freq - 760;
  ReadRegisters();
  Si4703_Registers[CHANNEL] &= 0xFC00;                              //Clear out the channel bits
  Si4703_Registers[CHANNEL] |= Channel;                             //Mask in the new channel
  Si4703_Registers[CHANNEL] |= 0x8000;                              //Set the TUNE bit to start
  UpdateRegisters();

  delay(60);                                                        //Seek/tune time is less than 60ms/channel. Silicon Laboratories recommends that all SDIO data traffic be suspended during Si4702/03-C19 seek and tune operations.
  while(1)                                                          //Poll to see if STC is set
  {
    ReadRegisters();
    if((Si4703_Registers[STATUSRSSI] & 0x4000) != 0)
    {
      break; //Tuning complete!
    }
  }

  ReadRegisters();
  Si4703_Registers[CHANNEL] &= 0x7FFF;                              //Clear the tune after a tune has completed
  UpdateRegisters();

  while(1)                                                          //Wait for the si4703 to clear the STC as well
  {
    ReadRegisters();
    if( (Si4703_Registers[STATUSRSSI] & 0x4000) == 0)
    {
      break;                                                        //Tuning complete!
    }
  }
}

void Si4703_Breakout::SetVolume(unsigned char Volume)               //0 to 15
{
  if((Volume < 0)||(Volume > 15))
  {
    return;
  }
  ReadRegisters();                                                  //Read the current register set
  Si4703_Registers[SYSCONFIG2] &= 0xFFF0;                           //Clear Volume bits
  Si4703_Registers[SYSCONFIG2] |= Volume;                           //Set new Volume
  UpdateRegisters();                                                //Update
}

unsigned short Si4703_Breakout::SeekUp()
{
  return Seek(SEEK_UP);
}

unsigned short Si4703_Breakout::SeekDown()
{
  return Seek(SEEK_DOWN);
}


void Si4703_Breakout::ReadRDS(unsigned char* Message, unsigned long Timeout)      //Message should be at least 9 chars. Result will be null terminated. Timeout in milliseconds
{ 
	long EndTime = millis() + Timeout;
  bool Completed[] = {false, false, false, false};
  unsigned short CompletedCount = 0;
  while(CompletedCount < 4 && millis() < EndTime)
  {
  	ReadRegisters();
  	if(Si4703_Registers[STATUSRSSI] & 0x8000)
  	{
  		// ls 2 bits of B determine the 4 letter pairs
  		// once we have a full set return
  		// if you get nothing after 20 readings return with empty string
  	  unsigned short b = Si4703_Registers[RDSB];
  	  unsigned short index = b & 0x03;
  	  if (! Completed[index] && b < 500)
  	  {
  		  Completed[index] = true;
  		  CompletedCount ++;
  	  	unsigned char Dh = (Si4703_Registers[RDSD] & 0xFF00) >> 8;
        unsigned char Dl = (Si4703_Registers[RDSD] & 0x00FF);
  		  Message[index * 2] = Dh;
  		  Message[index * 2 + 1] = Dl;
  		// Serial.print(Si4703_Registers[RDSD]); Serial.print(" ");
  		// Serial.print(index);Serial.print(" ");
  		// Serial.write(Dh);
  		// Serial.write(Dl);
  		// Serial.println();
      }
        delay(40);                                                                //Wait for the RDS bit to clear
  	}
  	else
    {
  	  delay(30);                                                                  //From AN230, using the polling method 40ms should be sufficient amount of time between checks
  	}
  }
	if (millis() >= EndTime)
  {
		Message[0] ='\0';
		return;
	}

  Message[8] = '\0';
}

//Read the entire register control set from 0x00 to 0x0F
void Si4703_Breakout::ReadRegisters()
{
  //Si4703 begins reading from register upper register of 0x0A and reads to 0x0F, then loops to 0x00.
  Wire.requestFrom(Si4703, 32);                                     //We want to read the entire register set from 0x0A to 0x00 = 32 bytes.

  while(Wire.available() < 32) ;                                    //Wait for 16 words/32 bytes to come back from slave I2C device
  //We may want some time-out error here

  //Remember, register 0x0A comes in first so we have to shuffle the array around a bit
  for(unsigned short x = 0x0A ; ; x++)
  { //Read in these 32 bytes
    if(x == 0x10)
    {
      x = 0;                                                        //Loop back to zero
    }
    Si4703_Registers[x] = Wire.read() << 8;
    Si4703_Registers[x] |= Wire.read();
    if(x == 0x09)
    {
      break;                                                        //We're done!
    }
  }
}

//Write the control registers (0x02 to 0x07) to the Si4703
//It's a little weird, you don't write an I2C addres
//The Si4703 assumes you are writing to 0x02 first, then increments
unsigned char Si4703_Breakout::UpdateRegisters()
{
  Wire.beginTransmission(Si4703);
  //A write command automatically begins with register 0x02 so no need to send a write-to address
  //First we send the 0x02 to 0x07 control registers
  //In general, we should not write to registers 0x08 and 0x09
  for(unsigned short RegSpot = 0x02 ; RegSpot < 0x08 ; RegSpot++)
  {
    unsigned char High_Byte = Si4703_Registers[RegSpot] >> 8;
    unsigned char Low_Byte = Si4703_Registers[RegSpot] & 0x00FF;

    Wire.write(High_Byte);                                          //Upper 8 bits
    Wire.write(Low_Byte);                                           //Lower 8 bits
  }

  //End this transmission
  unsigned char ack = Wire.endTransmission();
  if(ack != 0)
  { //We have a problem! 
    return(FAIL);
  }

  return(SUCCESS);
}

//Seeks out the next available station
//Returns the freq if it made it
//Returns zero if failed
unsigned short Si4703_Breakout::Seek(unsigned char SeekDirection)
{
  ReadRegisters();
  //Set seek mode wrap bit
  Si4703_Registers[POWERCFG] |= 0x0400;                                       //Allow wrap
  //Si4703_Registers[POWERCFG] &= 0xFBFF;                                     //Disallow wrap
  if(SeekDirection == SEEK_DOWN)
  {
    Si4703_Registers[POWERCFG] &= 0xFDFF;                                     //Seek down is the default upon reset
  }
  else
  {
    Si4703_Registers[POWERCFG] |= 0x0200;                                     //Set the bit to seek up
  }

  Si4703_Registers[POWERCFG] |= 0x0100;                                       //Start seek
  UpdateRegisters();                                                          //Seeking will now start

  //Poll to see if STC is set
  while(1)
  {
    delay(60);                                                                //Seek/tune time is less than 60ms/channel. Silicon Laboratories recommends that all SDIO data traffic be suspended during Si4702/03-C19 seek and tune operations.
    ReadRegisters();
    if((Si4703_Registers[STATUSRSSI] & 0x4000 != 0))
    {
      break;                                                                  //Tuning complete!
    }
  }

  ReadRegisters();
  unsigned short ValueSFBL = Si4703_Registers[STATUSRSSI] & 0x2000;           //Store the value of SF/BL
  Si4703_Registers[POWERCFG] &= 0xFEFF;                                       //Clear the SEEK bit after seek has completed. The SEEK bit must be set low to clear SF/BL
  UpdateRegisters();

  //Wait for the Si4703 to clear the STC as well. Setting the SEEK or TUNE bit low will clear STC.
  while(1)
  {
    ReadRegisters();
    if((Si4703_Registers[STATUSRSSI] & 0x4000) == 0)
    {
      break;
    }
  }

  if(ValueSFBL != 0)
  { //The bit was set indicating we hit a band limit or failed to find a station
    return(0);
  }
  return GetFrequency();
}

//Reads the current frequency
unsigned short Si4703_Breakout::GetFrequency()
{
  ReadRegisters();
  unsigned short Freq = (Si4703_Registers[READCHAN] & 0x03FF) + 760;
  return(Freq);
}

//Reads the current  RSSI signal strength, stereo status, and RDS ready bit
//RSSI (Received Signal Strength Indicator).
//RSSI is measured units of dBuV in 1 dB increments with a maximum of approximately
//75 dBuV. Si4702/03-C19 does not report RSSI levels greater than 75 dBuV.
//Reads Stereo status and RDS status
void Si4703_Breakout::GetStatus(unsigned char* RSSI, unsigned char* IsStereo, unsigned char* IsRDSR)
{
  ReadRegisters();
  if(RSSI != NULL)
  {
    *RSSI = Si4703_Registers[STATUSRSSI] & 0x00FF;                            //Mask out everything but the lower 8 bits
  }
  if(IsStereo != NULL)
  {
    *IsStereo = (Si4703_Registers[STATUSRSSI] & 0x0100) >> 8;
  }
  if(IsRDSR != NULL)
  {
    *IsRDSR = (Si4703_Registers[STATUSRSSI] & 0x8000) >> 8;
  }
}

